/*****************************************************************************
 *
 * FILE:	aarray.c
 * DESCRIPTION:	Wyvern: associated array library
 * DATE:	Thu, Oct 21 2004
 * UPDATED:	Thu, Mar 30 2006
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2004-2006 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2004-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: aarray.c,v 1.5 2006/03/30 04:47:26 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <errno.h>
#include "wa.h"

/******************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	AARRAY_TABLE_SIZE	(128)

/******************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void		aa_free(AArray);
static void		aa_free_value(AAValue *);
static AArray		aa_new_aarray(const char *, const char *);
static AAValue *	aa_new_value(const char *);
static AArray		aa_lookup(AArray *, const char *, const char *);
static AArray		aa_lookup_key(AArray *, const char *);
static unsigned int	aa_hash(const char *, size_t);

/******************************************************************************
 *
 *	Local variables definition
 *
 *****************************************************************************/
static size_t	aa_table_size = 0;
static size_t	aa_bitmask    = 0;

/******************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
AArray *
wa_new(size)
	size_t	size;	/* should be a power of 2 */
{
  AArray *	new;

  if (size == 0) {
    aa_table_size = AARRAY_TABLE_SIZE;
  }
  else if (size > 0) {
    aa_table_size = size;
  }
  else {
    errno = EINVAL;
    return NULL;
  }
  aa_bitmask = aa_table_size - 1;

  new = (AArray *)calloc(aa_table_size, sizeof(AArray));

  return new;
}

void
wa_free(aarray)
	AArray *	aarray;
{
  if (aarray != NULL) {
    register unsigned int	i;

    for (i = 0; i < aa_table_size; i++) {
      if (aarray[i] != NULL) {
	register AArray	ap;
	register AArray	ap_next;

	for (ap = aarray[i]; ap != NULL; ap = ap_next) {
	  ap_next = ap->next;
	  aa_free(ap);
	}
      }
    }
    free((AArray *)aarray);
  }
}

static void
aa_free(ap)
	register AArray	ap;
{
  if (ap != NULL) {
    if (ap->key != NULL) { free((char *)ap->key); }
    aa_free_value(ap->value);
    free((AArray *)ap);	ap = NULL;
  }
}

static void
aa_free_value(vlist)
	register AAValue *	vlist;
{
  register AAValue *	vp;
  register AAValue *	vp_next;

  for (vp = vlist; vp != NULL; vp = vp_next) {
    vp_next = vp->next;
    if (vp->name != NULL) { free((char *)vp->name); }
    free((AAValue *)vp);	vp = NULL;
  }
}

/*****************************************************************************/

const char *
wa_put(aarray, key, val)
	AArray *	aarray;
	const char *	key;
	const char *	val;
{
  if (aarray != NULL && key != NULL && val != NULL) {
    register AArray	ap;

    ap = aa_lookup(aarray, key, val);
    if (ap != NULL) {	/* replace value */
      if (ap->value != NULL) { aa_free_value(ap->value); }
      ap->value = aa_new_value(val);

      return ap->value != NULL ? ap->value->name : NULL;
    }
    else {
      unsigned int	hval;

      hval = aa_hash(key, strlen(key)) & aa_bitmask;
      ap   = aa_new_aarray(key, val);
      if (ap != NULL) {
	ap->next     = aarray[hval];
	aarray[hval] = ap;

	return ap->value != NULL ? ap->value->name : NULL;
      }
    }
  }
  else {
    errno = EINVAL;
  }

  return NULL;
}

AAValue *
wa_put_plus(aarray, key, val)
	AArray *	aarray;
	const char *	key;
	const char *	val;
{
  if (aarray != NULL && key != NULL && val != NULL) {
    register AArray	ap;

    ap = aa_lookup(aarray, key, val);
    if (ap == NULL) {
      unsigned int	hval;

      hval = aa_hash(key, strlen(key)) & aa_bitmask;
      ap   = aarray[hval];
      if (ap == NULL) {
	ap = aarray[hval] = aa_new_aarray(key, val);
      }
      else if (ap->count > 0 && strcmp(key, ap->key) == 0) {
	AAValue *	vp;

	vp = aa_new_value(val);
	if (vp != NULL) {
	  vp->next  = ap->value;
	  ap->value = vp;
	  ap->count++;
	}
      }
      else {
	AArray	new;

	new	     = aa_new_aarray(key, val);
	new->next    = ap;
	aarray[hval] = new;
      }
    }

    return ap->value;
  }
  else {
    errno = EINVAL;
  }

  return NULL;
}

static AArray
aa_new_aarray(key, val)
	const char *	key;
	const char *	val;
{
  AArray	new;

  new = (AArray)malloc(sizeof(AAHash));
  if (new != NULL) {
    new->key	= strdup(key);
    new->value	= aa_new_value(val);
    new->count	= 1;
    new->next	= NULL;
  }

  return new;
}

static AAValue *
aa_new_value(val)
	const char *	val;
{
  AAValue *	new;

  new = (AAValue *)malloc(sizeof(AAValue));
  if (new != NULL) {
    new->name = strdup(val);
    new->next = NULL;
  }

  return new;
}

/*****************************************************************************/

char *
wa_get_all(aarray, key, cs)
	AArray *	aarray;
	const char *	key;
	const char *	cs;	/* connection string */
{
#define	INIT_BUFSIZE	(80)
  char *	retval;
  size_t	cslen;
  size_t	total = INIT_BUFSIZE;
  
  if (cs == NULL) {
    errno = EINVAL;
    return NULL;
  }
  cslen = strlen(cs);

  retval = (char *)calloc(total, sizeof(char));
  if (retval != NULL) {
    register AAValue *	vp;
    size_t		sum = 0;

    for (vp = wa_get_plus(aarray, key); vp != NULL; vp = vp->next) {
      size_t	len;

      if (vp->name == NULL) { continue; }
      len = strlen(vp->name) + cslen;
      sum += len;
      if (sum >= total) {
	char *	wp;

	total	= sum + 1;
	wp	= (char *)realloc(retval, total);
	if (wp == NULL) {
	  free(retval);
	  return NULL;
	}
	retval	= wp;
      }
      len = strlcat(retval, vp->name, total);
      len = strlcat(retval, cs,       total);
    }
    retval[strlen(retval) - cslen] = '\0';
  }

  return retval;
}

const char *
wa_get(aarray, key)
	AArray *	aarray;
	const char *	key;
{
  register AAValue *	vp;

  vp = wa_get_plus(aarray, key);
  if (vp != NULL) {
    return vp->name;
  }

  return NULL;
}

AAValue *
wa_get_plus(aarray, key)
	AArray *	aarray;
	const char *	key;
{
  if (aarray != NULL && key != NULL) {
    register AArray	ap;

    ap = aa_lookup_key(aarray, key);
    if (ap != NULL) {
      return ap->value;
    }
    errno = ENOENT;
  }
  else {
    errno = EINVAL;
  }

  return NULL;
}

/*****************************************************************************/

int
wa_delete(aarray, key)
	AArray *	aarray;
	const char *	key;
{
  if (aarray != NULL && key != NULL) {
    unsigned int	h;
    register AArray	ap;
    register AArray	pp;	/* previous pointer */

    h = aa_hash(key, strlen(key)) & aa_bitmask;
    for (pp = ap = aarray[h]; ap != NULL; pp = ap, ap = ap->next) {
      if (ap->key != NULL && strcmp(key, ap->key) == 0) {
	if (pp == aarray[h]) {
	  aarray[h] = ap->next;
	}
	else {
	  pp->next = ap->next;
	}
	aa_free(ap);

	return 0;
      }
    }
  }
  errno = EINVAL;

  return -1;
}

int
wa_flush(aarray)	/* clear value for each key */
	AArray *	aarray;
{
  if (aarray != NULL) {
    register unsigned int	i;

    for (i = 0; i < aa_table_size; i++) {
      if (aarray[i] != NULL) {
	register AArray	ap;

	for (ap = aarray[i]; ap != NULL; ap = ap->next) {
	  aa_free_value(ap->value);	ap->value = NULL;
	}
      }
    }

    return 0;
  }
  errno = EINVAL;

  return -1;
}

/*****************************************************************************/

int
wa_foreach(aarray, action)
	AArray *	aarray;
	void		(*action)(const char *, const char *);
{
  if (aarray != NULL) {
    register unsigned int	i;

    for (i = 0; i < aa_table_size; i++) {
      if (aarray[i] != NULL) {
	register AArray	ap;

	for (ap = aarray[i]; ap != NULL; ap = ap->next) {
	  register AAValue *	vp;

	  for (vp = ap->value; vp != NULL; vp = vp->next) {
	    (*action)(ap->key, vp->name);
	  }
	}
      }
    }

    return 0;
  }
  errno = EINVAL;

  return -1;
}

/*****************************************************************************/

static AArray
aa_lookup(aarray, key, val)
	AArray *	aarray;
	const char *	key;
	const char *	val;
{
  if (aarray != NULL && key != NULL && val != NULL) {
    register AArray	ap;

    ap = aa_lookup_key(aarray, key);
    if (ap != NULL) {
      register AAValue *	vp;

      for (vp = ap->value; vp != NULL; vp = vp->next) {
       	if (vp->name != NULL && strcmp(val, vp->name) == 0) {
  	  return ap;
   	}
      }
    }
    errno = ENOENT;
  }
  else {
    errno = EINVAL;
  }

  return NULL;
}

static AArray
aa_lookup_key(aarray, key)
	AArray *	aarray;
	const char *	key;
{
  if (aarray != NULL && key != NULL) {
    unsigned int	hval;
    register AArray	ap;

    hval = aa_hash(key, strlen(key)) & aa_bitmask;
    for (ap = aarray[hval]; ap != NULL; ap = ap->next) {
      if (ap->key != NULL && strcmp(key, ap->key) == 0) {
	return ap;
      }
    }
    errno = ENOENT;
  }
  else {
    errno = EINVAL;
  }

  return NULL;
}

/*
 * A Hash Function for Hash Table Lookup
 * <http://burtleburtle.net/bob/hash/doobs.html>
 */
#undef getshort
#if (defined(__GNUC__) && defined(__i386__)) || defined(__WATCOMC__)
#define	getshort(d)	(*((const unsigned short *)(d)))
#endif
#if defined(_MSC_VER) || defined(__BORLANDC__)
#define	getshort(d)	(*((const unsigned short *)(d)))
#endif

#if !defined(getshort)
#define	getshort(d)	((((const unsigned char *)(d))[1] << 8UL) \
			+((const unsigned char *)(d))[0])
#endif
static unsigned int
aa_hash(data, len)
	const char *	data;
	size_t		len;
{
  unsigned long	hash;
  int		rem;
  register int	i;

  if (len <= 0 || data == NULL) {
    return 0;
  }

  hash = (unsigned long)len;	/* Avoid 0 -> 0 trap */
  rem  = len & 3;
  len >>= 2;

  /* Main loop */
  for (i = 0; i < len; i++) {
    hash += getshort(data);
    data += 2 * sizeof(char);
    hash ^= hash << 16;
    hash ^= getshort(data) << 11;
    data += 2 * sizeof(char);
    hash += hash >> 11;
  }

  /* Handle end cases */
  switch (rem) {
    case 3:
      hash += getshort(data);
      hash ^= hash << 16;
      hash ^= data[2 * sizeof(char)] << 18;
      hash += hash >> 11;
      break;
    case 2:
      hash += getshort(data);
      hash ^= hash << 11;
      hash += hash >> 17;
      break;
    case 1:
      hash += *data;
      hash ^= hash << 10;
      hash += hash >> 1;
      break;
    default:
      break;
  }

  /* Force "avalanching" of final 127 bits */
  hash ^= hash << 3;
  hash += hash >> 5;
  hash ^= hash << 2;
  hash += hash >> 15;
  hash ^= hash << 10;

  return hash;
}

/*****************************************************************************/

#if	DEBUG_AARRAY
static void	dump(AArray *);
static void	printer(const char *, const char *);

static void
dump(aarray)
	AArray *	aarray;
{
  if (aarray != NULL) {
    register unsigned int	i;

    for (i = 0; i < aa_table_size; i++) {
      if (aarray[i] != NULL) {
	register AArray	ap;

	for (ap = aarray[i]; ap != NULL; ap = ap->next) {
	  fprintf(stderr, "DEBUG[%02d] %s => %s\n",
		  i, ap->key, ap->value ? ap->value->name : "");
	}
      }
    }
  }
}

static void
printer(key, value)
	const char *	key;
	const char *	value;
{
  if (key != NULL && value != NULL) {
    fprintf(stdout, "%s: %s\n", key, value);
  }
}

int
main(argc, argv, envp)
	int	argc;
	char *	argv[];
	char *	envp[];
{
  AArray *	aa;

  aa = wa_new(WA_DEFAULT_SIZE);
  if (aa != NULL) {
    register char **	ep;

    for (ep = envp; *ep != NULL; ep++) {
      char *	p;
      char *	q = NULL;

      p = strchr(*ep, '=');
      if (p != NULL) {
	*p++ = '\0';
	q = p;
      }
      wa_put(aa, *ep, q);
    }
    dump(aa);
    wa_foreach(aa, printer);
    wa_delete(aa, "XMODIFIERS");
    wa_delete(aa, "USER");
    wa_delete(aa, "OSTYPE");
    dump(aa);
    wa_flush(aa);
    dump(aa);
    wa_free(aa);
  }

  return 0;
}
#endif	/* DEBUG_AARRAY */

