/*****************************************************************************
 *
 * FILE:	error.c
 * DESCRIPTION:	Wyvern: send error messages
 * DATE:	Sun, Sep 26 1999
 * UPDATED:	Thu, Feb  9 2006
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 1999-2006 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 1999-2006 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: error.c,v 1.20 2006/02/09 06:47:43 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#include "wyvern.h"
#include "mesg.h"

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static ErrorDoc *	lookup(const ErrorDoc *, const int);

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wv_add_error_documents
 * Description:		register response code and it's document file into hash
 * Arguments:		errorDocTable, response code, document location, type
 * Return Value:	none
 */
void
wv_add_error_documents(errorDocTable, code, text, type)
	ErrorDoc **	errorDocTable;
	const int	code;
	const char *	text;
	errdoc_t	type;
{
  /* 4xx: client error, 5xx: server error */
  if ((code >= 400 && code < 500) &&
      lookup(*errorDocTable, code) == NULL) {
    ErrorDoc *	new;

    new		   = (ErrorDoc *)ws_malloc(sizeof(ErrorDoc));
    new->code	   = code;
    new->text	   = ws_strdup(text);
    new->type	   = type;
    new->next	   = *errorDocTable;
    *errorDocTable = new;
  }
}

/*
 * Function Name:	lookup
 * Description:		lookup the error document with response status code
 * Arguments:		errorDocTable, response status code
 * Return Value:	pointer of error document
 */
static ErrorDoc *
lookup(errorDocTable, code)
	const ErrorDoc *	errorDocTable;
	register const int	code;
{
  register ErrorDoc *	edp;

  for (edp = (ErrorDoc *)errorDocTable; edp != NULL; edp = edp->next) {
    if (edp->code == code) {
      return edp;
    }
  }

  return NULL;
}

/*****************************************************************************/

/*
 * Function Name:	wv_error_doc_checker
 * Description:		customizable error document performer
 * Arguments:		message
 * Return Value:	none
 */
void
wv_error_doc_checker(msg)
	register Message *	msg;
{
  register ErrorDoc *	edp;

  DBG(fprintf(stderr, "[errorDocChecker] status(%d)\n",
	      wvResponseStatusTable[msg->response.status].code);)
  if (msg->response.status <= STATUS_CODE_4XX) {
    return;	/* no error */
  }

  /* Close opened file to response */
  if (msg->file.nfd != NULL) {
    msg->close(msg->file.nfd);
    msg->file.nfd = NULL;
  }

  /* Lookup for user defined error document */
  edp = lookup(msg->vhost->errorDocTable,
	       wvResponseStatusTable[msg->response.status].code);
  if (edp != NULL) {
    switch (edp->type) {
      case ED_STRING:
      	msg->response.error_mesg = edp->text;
       	break;
      case ED_FILE: {
  	  char *	path;
  	  int		fd;

  	  ws_asprintf(&path, "%s%s", msg->vhost->documentRoot, edp->text);
  	  fd = ws_safe_open(path, O_RDONLY, &msg->file.status, false);
  	  free((char *)path);
  	  if (fd != -1) {
  	    msg->file.nfd = wt_netfd_fdopen(fd);
  	  }
   	}
    	break;
      case ED_URL:
      	wa_put(msg->response.header, "Location", edp->text);
       	msg->response.status = STATUS_MOVED_PERMANENTLY;
      default:
	break;
    }
  }
}

/*****************************************************************************/

/*
 * Function Name:	wv_error_message_sender
 * Description:		send the error message
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
wv_error_message_sender(msg)
	register Message *	msg;
{
  char		buf[HUGE_BUFSIZE];
  size_t	size = sizeof(buf);
  char *	bufp = buf;
  size_t	len;

  memset(buf, 0, sizeof(buf));
  len = snprintf(bufp, size,
	"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n"
	"<html lang=\"ja\">\n"
	"<head>\n"
	"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=EUC-JP\">\n"
	"<style type=\"text/css\">\n"
	"<!--\n"
	"body { background: white }\n"
	"h1 { text-align: center; color: red }\n"
	"-->\n"
	"</style>\n"
	"<title>%d %s</title>\n"
	"</head>\n"
	"<body lang=\"ja\">\n"
	"<h1>%d %s</h1>\n"
	"<p style=\"text-align: center\">\n",
	wvResponseStatusTable[msg->response.status].code,
	wvResponseStatusTable[msg->response.status].phrase,
	wvResponseStatusTable[msg->response.status].code,
	wvResponseStatusTable[msg->response.status].phrase);
  bufp += len;
  size -= len;

  if (msg->response.error_mesg == NULL) {
    char *	uri = ws_sanitize_uri(msg->request.uri);

    switch (msg->response.status) {
      case STATUS_BAD_REQUEST:			/* 400 */
	len = snprintf(bufp, size, "%s", MSG_BAD_REQUEST);
	break;
      case STATUS_UNAUTHORIZED:			/* 401 */
	len = snprintf(bufp, size, MSG_UNAUTHORIZED, uri);
	break;
      case STATUS_FORBIDDEN:			/* 403 */
	len = snprintf(bufp, size, MSG_FORBIDDEN, uri);
	break;
      case STATUS_NOT_FOUND:			/* 404 */
	len = snprintf(bufp, size, MSG_NOT_FOUND, uri);
	break;
      case STATUS_METHOD_NOT_ALLOWED:		/* 405 */
	len = snprintf(bufp, size, MSG_METHOD_NOT_ALLOWED,
		       msg->request.method.name);
	break;
      case STATUS_NOT_ACCEPTABLE:		/* 406 */
	len = snprintf(bufp, size, MSG_NOT_ACCEPTABLE, msg->file.mime);
	break;
      case STATUS_REQUEST_TIMEOUT:		/* 408 */
	len = snprintf(bufp, size, "%s", MSG_REQUEST_TIMEOUT);
	break;
      case STATUS_PRECONDITION_FAILED:		/* 412 */
	len = snprintf(bufp, size, "%s", MSG_PRECONDITION_FAILED);
	break;
      case STATUS_REQUEST_URI_TOO_LONG:		/* 414 */
	len = snprintf(bufp, size, "%s", MSG_REQUEST_URI_TOO_LONG);
	break;
      case STATUS_INTERNAL_SERVER_ERROR:	/* 500 */
	len = snprintf(bufp, size, MSG_INTERNAL_SERVER_ERROR, uri);
	break;
      case STATUS_NOT_IMPLEMENTED:		/* 501 */
	len = snprintf(bufp, size, MSG_NOT_IMPLEMENTED, uri);
	break;
      case STATUS_GATEWAY_TIMEOUT:		/* 504 */
	len = snprintf(bufp, size, "%s", MSG_GATEWAY_TIMEOUT);
	break;
      case STATUS_HTTP_VERSION_NOT_SUPPORTED:	/* 505 */
	len = snprintf(bufp, size, "%s", MSG_HTTP_VERSION_NOT_SUPPORTED);
	break;
      default:
	break;
    }
    wv_free((char *)uri);
  }
  else {
    len = snprintf(bufp, size, "%s", msg->response.error_mesg);
  }
  bufp += len;
  size -= len;

  len = snprintf(bufp, size,
	"</p>\n"
	"<hr>\n"
	"<address>\n"
	"Powered by&nbsp;"
	"<a href=\"%s\">%s</a>\n"
	"&nbsp;&lt;<a href=\"mailto:%s\">%s</a>&gt;\n"
	"</address>\n"
	"</body>\n"
	"</html>\n",
	SERVER_ADDRESS, SERVER_VERSION,
	msg->vhost->serverAdmin, msg->vhost->serverAdmin);

  /* Send error message */
  msg->write(msg->nsd, buf, strlen(buf), WT_UTIME_NO_TIMEOUT);

  return 0;
}

